"""
Copyright 2017 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import logging
import os
import shutil
import tempfile
import unittest
from collections.abc import Mapping
from importlib.abc import Loader
from typing import Optional
from unittest import mock

import py
import pytest

from _io import StringIO
from inmanta import const, env, module
from inmanta.ast import CompilerException
from inmanta.compiler.help.explainer import ExplainerFactory
from inmanta.data.model import PipConfig
from inmanta.env import LocalPackagePath
from inmanta.loader import PluginModuleFinder, PluginModuleLoader
from inmanta.module import InmantaModuleRequirement
from inmanta.moduletool import ModuleTool
from utils import module_from_template


def test_module():
    good_mod_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data", "modules", "mod1")
    module.ModuleV1(project=mock.Mock(), path=good_mod_dir)


def test_bad_module():
    bad_mod_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data", "modules", "mod2")
    with pytest.raises(module.ModuleMetadataFileNotFound):
        module.ModuleV1(project=mock.Mock(), path=bad_mod_dir)


class TestModuleName(unittest.TestCase):
    def __init__(self, methodName="runTest"):  # noqa: N803
        unittest.TestCase.__init__(self, methodName)

        self.stream = None
        self.handler = None
        self.log = None

    def setUp(self):
        self.stream = StringIO()
        self.handler = logging.StreamHandler(self.stream)
        self.log = logging.getLogger(module.__name__)

        for handler in self.log.handlers:
            self.log.removeHandler(handler)

        self.log.addHandler(self.handler)

    def test_wrong_name(self):
        mod_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data", "modules", "mod3")
        module.ModuleV1(project=mock.Mock(), path=mod_dir)

        self.handler.flush()
        assert "The name in the module file (mod1) does not match the directory name (mod3)" in self.stream.getvalue().strip()

    def test_non_matching_name_v2_module(self) -> None:
        """
        Make sure the warning regarding directory name does not trigger for v2 modules, as it is not relevant there.
        """
        template_dir: str = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data", "modules_v2", "minimalv2module")
        with tempfile.TemporaryDirectory() as tmpdir:
            mod_dir: str = os.path.join(tmpdir, "not-the-module-name")
            module_from_template(template_dir, mod_dir)
            module.ModuleV2(project=module.DummyProject(), path=mod_dir)
            self.handler.flush()
            assert self.stream.getvalue().strip() == ""

    def tearDown(self):
        self.log.removeHandler(self.handler)
        self.handler.close()


def test_to_v2():
    """
    Test whether the `to_v2()` method of `ModuleV1Metadata` works correctly.
    """
    v1_metadata = module.ModuleV1Metadata(
        name="a_test_module",
        description="A description",
        version="1.2.3",
        license="Apache 2.0",
        compiler_version="4.5.6",
        requires=["module_dep_1", "module_dep_2"],
    )
    v2_metadata = v1_metadata.to_v2()
    for attr_name in ["description", "version", "license"]:
        assert v1_metadata.__getattribute__(attr_name) == v2_metadata.__getattribute__(attr_name)

    def _convert_module_to_package_name(module_name: str) -> str:
        return f"{module.ModuleV2.PKG_NAME_PREFIX}{module_name.replace('_', '-')}"

    assert _convert_module_to_package_name(v1_metadata.name) == v2_metadata.name
    assert [_convert_module_to_package_name(req) for req in v1_metadata.requires] == v2_metadata.install_requires


@pytest.mark.slowtest
def test_is_versioned(
    snippetcompiler_clean,
    modules_dir: str,
    modules_v2_dir: str,
    caplog,
    tmpdir,
    local_module_package_index,  # upstream for setuptools for isolated build
) -> None:
    """
    Test whether the warning regarding non-versioned modules is given correctly.
    """
    # Disable modules_dir
    snippetcompiler_clean.modules_dir = None

    def compile_and_assert_warning(
        module_name: str, needs_versioning_warning: bool, install_v2_modules: list[LocalPackagePath] = []
    ) -> None:
        caplog.clear()
        snippetcompiler_clean.setup_for_snippet(
            f"import {module_name}", autostd=False, install_v2_modules=install_v2_modules, index_url=local_module_package_index
        )
        snippetcompiler_clean.do_export()
        warning_message = f"Module {module_name} is not version controlled, we recommend you do this as soon as possible."
        assert (warning_message in caplog.text) is needs_versioning_warning

    # V1 module
    module_name_v1 = "mod1"
    module_dir = os.path.join(modules_dir, module_name_v1)
    module_copy_dir = os.path.join(snippetcompiler_clean.libs, module_name_v1)
    shutil.copytree(module_dir, module_copy_dir)
    dot_git_dir = os.path.join(module_copy_dir, ".git")
    assert not os.path.exists(dot_git_dir)
    compile_and_assert_warning(module_name_v1, needs_versioning_warning=True)
    os.mkdir(dot_git_dir)
    compile_and_assert_warning(module_name_v1, needs_versioning_warning=False)

    # V2 module
    module_name_v2 = "elaboratev2module"
    module_dir = os.path.join(modules_v2_dir, module_name_v2)
    module_copy_dir = os.path.join(tmpdir, module_name_v2)
    shutil.copytree(module_dir, module_copy_dir)
    dot_git_dir = os.path.join(module_copy_dir, ".git")
    assert not os.path.exists(dot_git_dir)
    # Non-editable install can never be checked for versioning
    compile_and_assert_warning(
        module_name_v2,
        needs_versioning_warning=False,
        install_v2_modules=[LocalPackagePath(path=module_copy_dir, editable=False)],
    )
    compile_and_assert_warning(
        module_name_v2,
        needs_versioning_warning=True,
        install_v2_modules=[LocalPackagePath(path=module_copy_dir, editable=True)],
    )
    os.mkdir(dot_git_dir)
    # Non-editable install can never be checked for versioning
    compile_and_assert_warning(
        module_name_v2,
        needs_versioning_warning=False,
        install_v2_modules=[LocalPackagePath(path=module_copy_dir, editable=False)],
    )
    compile_and_assert_warning(
        module_name_v2,
        needs_versioning_warning=False,
        install_v2_modules=[LocalPackagePath(path=module_copy_dir, editable=True)],
    )


@pytest.mark.parametrize(
    "v1_module, all_python_requirements,strict_python_requirements,module_requirements,module_v2_requirements",
    [
        (
            True,
            ["jinja2~=3.2.1", "inmanta-module-v2-module==1.2.3"],
            ["jinja2~=3.2.1"],
            ["v2_module==1.2.3", "v1_module==1.1.1"],
            [InmantaModuleRequirement.parse("v2_module==1.2.3")],
        ),
        (
            False,
            ["jinja2~=3.2.1", "inmanta-module-v2-module==1.2.3"],
            ["jinja2~=3.2.1"],
            ["v2_module==1.2.3"],
            [InmantaModuleRequirement.parse("v2_module==1.2.3")],
        ),
    ],
)
def test_get_requirements(
    modules_dir: str,
    modules_v2_dir: str,
    v1_module: bool,
    all_python_requirements: list[str],
    strict_python_requirements: list[str],
    module_requirements: list[str],
    module_v2_requirements: list[str],
) -> None:
    """
    Test the different methods to get the requirements of a module.
    """
    module_name = "many_dependencies"

    if v1_module:
        module_dir = os.path.join(modules_dir, module_name)
        mod = module.ModuleV1(module.DummyProject(autostd=False), module_dir)
    else:
        module_dir = os.path.join(modules_v2_dir, module_name)
        mod = module.ModuleV2(module.DummyProject(autostd=False), module_dir)

    assert set(mod.get_all_python_requirements_as_list()) == set(all_python_requirements)
    assert set(mod.get_strict_python_requirements_as_list()) == set(strict_python_requirements)
    assert set(mod.get_module_requirements()) == set(module_requirements)
    assert set(mod.get_module_v2_requirements()) == set(module_v2_requirements)
    assert set(mod.requires()) == {module.InmantaModuleRequirement.parse(req) for req in module_requirements}


@pytest.mark.parametrize("editable", [True, False])
def test_module_v2_source_get_installed_module_editable(
    # Use clean snippetcompiler (separate venv) because this test installs test packages into the snippetcompiler venv.
    snippetcompiler_clean,
    modules_v2_dir: str,
    editable: bool,
    local_module_package_index,  # upstream for setuptools for isolated build
) -> None:
    """
    Make sure ModuleV2Source.get_installed_module identifies editable installations correctly.
    """
    module_name: str = "minimalv2module"
    module_dir: str = os.path.join(modules_v2_dir, module_name)
    snippetcompiler_clean.setup_for_snippet(
        f"import {module_name}",
        autostd=False,
        install_v2_modules=[env.LocalPackagePath(path=module_dir, editable=editable)],
        index_url=local_module_package_index if editable else None,
    )

    source: module.ModuleV2Source = module.ModuleV2Source()
    mod: Optional[module.ModuleV2] = source.get_installed_module(module.DummyProject(autostd=False), module_name)
    assert mod is not None
    # os.path.realpath because snippetcompiler uses symlinks
    assert os.path.realpath(mod.path) == (
        module_dir if editable else os.path.join(env.process_env.site_packages_dir, "inmanta_plugins", module_name)
    )
    assert mod._is_editable_install == editable


def test_module_v2_source_path_for_v1(snippetcompiler) -> None:
    """
    Make sure ModuleV2Source.path_for does not include modules loaded by the v1 module loader.
    """
    # load tests module
    snippetcompiler.setup_for_snippet("import tests")
    module.Project.get().load_plugins()

    # make sure the v1 module finder is configured and discovered by env.process_env
    assert PluginModuleFinder.MODULE_FINDER is not None
    module_info: Optional[tuple[Optional[str], Loader]] = env.process_env.get_module_file("inmanta_plugins.tests")
    assert module_info is not None
    path, loader = module_info
    assert path is not None
    assert isinstance(loader, PluginModuleLoader)

    source: module.ModuleV2Source = module.ModuleV2Source()
    assert source.path_for("tests") is None


def test_module_v2_from_v1_path(
    local_module_package_index: str,
    modules_v2_dir: str,
    snippetcompiler_clean,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """
    Verify that attempting to load a v2 module from the v1 modules path fails with an appropriate message when a v2 module is
    found in a v1 module source.
    """
    with pytest.raises(module.ModuleLoadingException) as excinfo:
        snippetcompiler_clean.setup_for_snippet("import minimalv2module", add_to_module_path=[modules_v2_dir])
    cause: CompilerException = excinfo.value.__cause__
    assert cause.msg == (
        "Module at %s looks like a v2 module. Please have a look at the documentation on how to use v2 modules."
        % os.path.join(modules_v2_dir, "minimalv2module")
    )
    assert (
        ExplainerFactory().explain_and_format(excinfo.value, plain=True).strip()
        == (
            f"""
Exception explanation
=====================
This error occurs when a v2 module was found in v1 modules path. To resolve this you should either convert this module to be v1 or install it as a v2 module and set up your project accordingly.

If you want to use the module as a v1 module, make sure to use the v1 cookiecutter template to create new modules.

If you want to use the module as a v2 module:
- set up your project with a module source of type "package" (see documentation)
- if you would like to work in editable mode on a local copy of the module, run `inmanta module install -e {modules_v2_dir}/minimalv2module`
- run `inmanta module add --v2 minimalv2module` to add the module as a dependency and to install the module if required.
        """  # noqa: E501
        ).strip()
    )

    # verify that adding it as a v2 resolves the issue
    project: module.Project = snippetcompiler_clean.setup_for_snippet(
        "",
        autostd=False,
        index_url=local_module_package_index,
        install_project=False,
    )
    os.chdir(project.path)
    project_init = module.Project.__init__

    # patch Project.__init__ to set autostd=False because v2 std does not exist yet
    def project_init_nostd(self, *args, **kwargs) -> None:
        project_init(self, *args, autostd=False, **{key: value for key, value in kwargs.items() if key != "autostd"})

    monkeypatch.setattr(module.Project, "__init__", project_init_nostd)
    ModuleTool().add("minimalv2module", v2=True)
    snippetcompiler_clean.setup_for_snippet(
        "import minimalv2module",
        autostd=False,
        # run with the same module path
        add_to_module_path=[modules_v2_dir],
        python_package_sources=[local_module_package_index],
        install_project=True,
    )


@pytest.mark.slowtest
def test_module_v2_incorrect_install_warning(
    tmpdir: py.path.local,
    modules_v2_dir: str,
    snippetcompiler_clean,
    caplog,
) -> None:
    """
    Verify that attempting to load a v2 module that has been installed from source with `pip install` rather than
    `inmanta module install` results in an appropriate error or warning.
    """
    # set up project and activate project venv
    snippetcompiler_clean.setup_for_snippet("")

    # prepare module
    module_dir: str = str(tmpdir.join("mymodule"))
    shutil.copytree(os.path.join(modules_v2_dir, "minimalv2module"), module_dir)

    def verify_exception(expected: Optional[str]) -> None:
        """
        Verify AST loading fails with the expected message, or succeeds if expected is None.
        """
        if expected is None:
            snippetcompiler_clean.setup_for_snippet("import minimalv2module", autostd=False)
            return
        with pytest.raises(module.ModuleLoadingException) as excinfo:
            snippetcompiler_clean.setup_for_snippet("import minimalv2module", autostd=False)
        cause: CompilerException = excinfo.value.__cause__
        assert cause.msg == expected

    # install module from source without using `inmanta module install`
    env.process_env.install_for_config(
        requirements=[], paths=[env.LocalPackagePath(path=module_dir, editable=False)], config=PipConfig(use_system_config=True)
    )
    module_path = os.path.join(env.process_env.site_packages_dir, const.PLUGINS_PACKAGE, "minimalv2module")
    verify_exception(
        f"Invalid module at {module_path}: found module package but it has no setup.cfg. "
        "This occurs when you install or build modules from source incorrectly. "
        "Always use the `inmanta module build` command followed by `pip install ./dist/<dist-package>` to "
        "respectively build a module from source and install the distribution "
        "package. Make sure to uninstall the broken package first."
    )

    # include setup.cfg in package to circumvent error
    shutil.copy(os.path.join(module_dir, "setup.cfg"), os.path.join(module_dir, const.PLUGINS_PACKAGE, "minimalv2module"))
    env.process_env.install_for_config(
        requirements=[], paths=[env.LocalPackagePath(path=module_dir, editable=False)], config=PipConfig(use_system_config=True)
    )
    verify_exception(
        "The module at %s contains no _init.cf file. This occurs when you install or build modules from source"
        " incorrectly. Always use the `inmanta module build` command followed by `pip install ./dist/<dist-package>` to"
        " respectively build a module from source and install the distribution package."
        " Make sure to uninstall the broken package first." % module_path
    )
    os.remove(os.path.join(module_dir, const.PLUGINS_PACKAGE, "minimalv2module", "setup.cfg"))

    # verify that proposed solution works: editable install doesn't require uninstall first
    env.process_env.install_for_config(
        requirements=[],
        paths=[env.LocalPackagePath(path=module_dir, editable=True)],
        config=PipConfig(use_system_config=True),
    )
    verify_exception(None)


def test_from_path(tmpdir: py.path.local, projects_dir: str, modules_dir: str, modules_v2_dir: str) -> None:
    """
    Verify that ModuleLike.from_path() and subclass overrides work as expected.
    """

    def check(
        path: str,
        *,
        subdir: Optional[str] = None,
        expected: Mapping[type[module.ModuleLike], Optional[type[module.ModuleLike]]],
    ) -> None:
        """
        Check the functionality for the given path and expected outcomes.

        :param path: The path to the root of the module like directory.
        :param subdir: The subpath to pass to `from_path`, relative to `path`.
        :param expected: Key-value pairs where keys represent the classes to call the method on and values the expected type
            for the return value.
        """
        full_path: str = os.path.join(path, *([subdir] if subdir is not None else []))
        for cls, tp in expected.items():
            result: Optional[module.ModuleLike] = cls.from_path(full_path)
            assert (result is None) is (tp is None)
            if tp is not None:
                assert isinstance(result, tp)
                assert result.path == path

    # project checks
    check(
        os.path.join(projects_dir, "simple_project"),
        expected={
            module.ModuleLike: module.Project,
            module.Project: module.Project,
            module.Module: None,
            module.ModuleV1: None,
            module.ModuleV2: None,
        },
    )

    # module v1 checks
    check(
        os.path.join(modules_dir, "minimalv1module"),
        expected={
            module.ModuleLike: module.ModuleV1,
            module.Project: None,
            module.Module: module.ModuleV1,
            module.ModuleV1: module.ModuleV1,
            module.ModuleV2: None,
        },
    )

    # module v2 checks
    check(
        os.path.join(modules_v2_dir, "minimalv2module"),
        expected={
            module.ModuleLike: module.ModuleV2,
            module.Project: None,
            module.Module: module.ModuleV2,
            module.ModuleV1: None,
            module.ModuleV2: module.ModuleV2,
        },
    )

    check(str(tmpdir), expected={module.ModuleLike: None})

    # advanced setup: project with modules in libs dir
    project_dir: str = str(tmpdir.join("project"))
    shutil.copytree(os.path.join(projects_dir, "simple_project"), project_dir)
    libs_dir: str = os.path.join(project_dir, "libs")
    os.makedirs(libs_dir, exist_ok=True)
    module_dir: str = os.path.join(libs_dir, "minimalv1module")
    shutil.copytree(os.path.join(modules_dir, "minimalv1module"), module_dir)
    check(module_dir, expected={module.ModuleLike: module.ModuleV1})
    check(libs_dir, expected={module.ModuleLike: None})
    check(project_dir, expected={module.ModuleLike: module.Project})



"""
Copyright 2019 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import hashlib
import importlib.abc
import importlib.machinery
import importlib.util
import inspect
import os
import shutil
import sys
from logging import DEBUG
from types import ModuleType
from typing import Optional

import py
import pytest
from pytest import fixture

import utils
from inmanta import const, env, loader, moduletool
from inmanta.env import PipConfig
from inmanta.loader import ModuleSource, SourceInfo
from inmanta.module import Project


def get_module_source(module: str, code: str) -> ModuleSource:
    data = code.encode()
    sha1sum = hashlib.new("sha1")
    sha1sum.update(data)
    hv: str = sha1sum.hexdigest()
    return ModuleSource(module, hv, False, data)


def test_code_manager(tmpdir: py.path.local, deactive_venv):
    """Verify the code manager"""
    original_project_dir: str = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data", "plugins_project")
    project_dir = os.path.join(tmpdir, "plugins_project")
    shutil.copytree(original_project_dir, project_dir)
    project: Project = Project(project_dir, venv_path=os.path.join(project_dir, ".env"))
    Project.set(project)
    project.install_modules()
    project.load()

    project.load_module("single_plugin_file", allow_v1=True)
    project.load_module("multiple_plugin_files", allow_v1=True)
    import inmanta_plugins.multiple_plugin_files.handlers as multi
    import inmanta_plugins.single_plugin_file as single

    mgr = loader.CodeManager()
    mgr.register_code("std::testing::NullResource", single.MyHandler)
    mgr.register_code("multiple_plugin_files::NullResourceBis", multi.MyHandler)

    def assert_content(source_info: SourceInfo, handler) -> str:
        filename = inspect.getsourcefile(handler)
        assert filename is not None
        content: str
        with open(filename, "rb") as fd:
            content = fd.read()
            assert source_info.content == content
            assert len(source_info.hash) > 0
            return content

    # get types
    types = dict(mgr.get_types())
    assert "std::testing::NullResource" in types
    assert "multiple_plugin_files::NullResourceBis" in types

    single_type_list: list[SourceInfo] = types["std::testing::NullResource"]
    multi_type_list: list[SourceInfo] = types["multiple_plugin_files::NullResourceBis"]

    assert len(single_type_list) == 1
    single_content: str = assert_content(single_type_list[0], single.MyHandler)

    assert len(multi_type_list) == 3
    multi_content: str = assert_content(
        next(s for s in multi_type_list if s.module_name == "inmanta_plugins.multiple_plugin_files.handlers"), multi.MyHandler
    )

    # get_file_hashes
    mgr_contents: set[bytes] = {mgr.get_file_content(hash) for hash in mgr.get_file_hashes()}
    assert single_content in mgr_contents
    assert multi_content in mgr_contents

    with pytest.raises(KeyError):
        mgr.get_file_content("test")

    # register type without source
    with pytest.raises(loader.SourceNotFoundException):
        mgr.register_code("test2", str)

    # verify requirements behavior
    source_info: SourceInfo = single_type_list[0]
    # by default also install dependencies on other modules
    assert source_info.requires == ["inmanta-module-std", "lorem"]
    project._metadata.agent_install_dependency_modules = False
    # reset cache
    source_info._requires = None
    # when disabled only install non-module dependencies
    assert source_info.requires == ["lorem"]


def test_code_loader(tmp_path, caplog):
    """
    Test code loader capabilities:
        - test code loader cache
        - test that an exception is raised when re-loading a module with different content
    """
    caplog.set_level(DEBUG)

    cl = loader.CodeLoader(tmp_path)

    with pytest.raises(ImportError):
        import inmanta_plugins.inmanta_unit_test  # NOQA

    code = """
def test():
    return 10
    """
    source_1 = get_module_source("inmanta_plugins.inmanta_unit_test", code)

    # Ensure source is present on disk
    cl.deploy_version([source_1])

    assert any("Deploying code " in message for message in caplog.messages)
    caplog.clear()

    # First manual load to be able to check the content remains untouched
    import inmanta_plugins.inmanta_unit_test  # NOQA

    assert inmanta_plugins.inmanta_unit_test.test() == 10

    # deploy same version
    cl.deploy_version([source_1])

    assert inmanta_plugins.inmanta_unit_test.test() == 10
    assert any("Deploying code " in message for message in caplog.messages)
    assert any(
        f"Not deploying code (hv={source_1.hash_value}, module={source_1.name}) because it is already on disk" in message
        for message in caplog.messages
    )
    caplog.clear()

    # Load the module to register it in the loader cache
    cl.load_module(source_1.name, source_1.hash_value)
    # Subsequent deploys of the same module will result in a cache hit
    cl.deploy_version([source_1])
    assert any(
        f"Not deploying code (hv={source_1.hash_value}, module={source_1.name}) because of cache hit" in message
        for message in caplog.messages
    )
    caplog.clear()

    # deploy new version
    code = """
def test():
    return 20
        """
    source_2 = get_module_source("inmanta_plugins.inmanta_unit_test", code)
    cl.deploy_version([source_2])

    assert any("Deploying code " in message for message in caplog.messages)

    with pytest.raises(Exception):
        cl.load_module(source_2.name, source_2.hash_value)
        assert any(
            f"The content of module {source_2.name} changed since it was last imported." in message
            for message in caplog.messages
        )

    assert inmanta_plugins.inmanta_unit_test.test() == 10


def test_code_loader_dependency(tmp_path, caplog, deactive_venv):
    """Test loading two modules with a dependency between them"""
    cl = loader.CodeLoader(tmp_path)

    source_init: ModuleSource = get_module_source(
        "inmanta_plugins.inmanta_unit_test_modular",
        """
        """,
    )

    source_tests: ModuleSource = get_module_source(
        "inmanta_plugins.inmanta_unit_test_modular.tests",
        """
from inmanta_plugins.inmanta_unit_test_modular.helpers import helper

def test():
    return 10 + helper()
        """,
    )

    source_helpers: ModuleSource = get_module_source(
        "inmanta_plugins.inmanta_unit_test_modular.helpers",
        """
def helper():
    return 1
        """,
    )

    cl.deploy_version([source_tests, source_helpers, source_init])

    import inmanta_plugins.inmanta_unit_test_modular.tests  # NOQA

    assert inmanta_plugins.inmanta_unit_test_modular.tests.test() == 11
    assert "ModuleNotFoundError: No module named" not in caplog.text


def test_2312_code_loader_missing_init(tmp_path) -> None:
    cl = loader.CodeLoader(tmp_path)

    code = """
def test():
    return 10
        """
    cl.deploy_version([get_module_source("inmanta_plugins.my_module.my_sub_mod", code)])

    import inmanta_plugins.my_module.my_sub_mod as sm

    assert sm.test() == 10


def test_code_loader_import_error(tmp_path, caplog, deactive_venv):
    """Test loading code with an import error"""
    cl = loader.CodeLoader(tmp_path)
    code = """
import badimmport
def test():
    return 10
    """

    with pytest.raises(ImportError):
        import inmanta_plugins.inmanta_bad_unit_test  # NOQA

    caplog.clear()
    cl.deploy_version([get_module_source("inmanta_plugins.inmanta_bad_unit_test", code)])

    with pytest.raises(ModuleNotFoundError):
        import inmanta_plugins.inmanta_bad_unit_test  # NOQA

        assert "ModuleNotFoundError: No module named 'badimmport'" in caplog.text


@fixture(scope="function")
def module_path(tmpdir):
    """
    Remark: Don't use this fixture in combination with instances of the Project class
            or the CodeLoader class as it would reset the config of the PluginModuleFinder.
    """
    loader.PluginModuleFinder.configure_module_finder(modulepaths=[str(tmpdir)])
    yield str(tmpdir)
    loader.PluginModuleFinder.reset()


@pytest.mark.slowtest
@pytest.mark.parametrize(
    "prefer_finder, reload",
    [(True, False), (False, False), (True, True)],
)
def test_plugin_module_finder(
    tmpdir: py.path.local,
    tmpvenv_active_inherit: env.VirtualEnv,
    modules_dir: str,
    prefer_finder: bool,
    reload: bool,
) -> None:
    """
    Verify correct behavior of the PluginModuleFinder class, especially with respect to preference when a module is present
    both in the normal venv and in the finder's module path.
    The different scenarios are tested via parametrization rather than in a single test case to force proper cleanup in
    between.

    :param prefer_finder: Configure the custom module finder to be preferred over the default finders.
    :param reload: Instead of only importing the module at the end, already import it before setting up the finder and reload
        it after, checking that the change of source works as expected.
    """
    module: str = "mymodule"
    python_module: str = f"{const.PLUGINS_PACKAGE}.{module}"

    # set up libs dir for the custom module finder
    libs_dir: py.path.local = tmpdir.mkdir("libs")
    module_dir: py.path.local = libs_dir.join(module)
    utils.v1_module_from_template(
        os.path.join(modules_dir, "minimalv1module"),
        str(module_dir),
        new_name=module,
        new_content_init_py="where = 'libs'",
    )

    # install module in venv
    venv_module_dir: py.path.local = tmpdir.join("mymodule_for_venv")
    utils.v1_module_from_template(
        str(module_dir),
        str(venv_module_dir),
        new_content_init_py="where = 'venv'",
    )
    mod_artifact_paths = moduletool.ModuleTool().build(path=str(venv_module_dir), wheel=True)
    env.process_env.install_for_config(
        requirements=[],
        paths=[env.LocalPackagePath(path=mod_artifact_paths[0])],
        config=PipConfig(use_system_config=True),
    )

    module_to_reload: Optional[ModuleType] = None
    if reload:
        # load it once before setting up the finder
        module_to_reload = importlib.import_module(python_module)

    # set up module finder
    assert not any(isinstance(finder, loader.PluginModuleFinder) for finder in sys.meta_path)
    loader.PluginModuleFinder.configure_module_finder(modulepaths=[str(libs_dir)], prefer=prefer_finder)

    # verify that the correct module will be loaded: either the one from the venv or the one in libs, depending on parameters
    assert isinstance(sys.meta_path[0 if prefer_finder else -1], loader.PluginModuleFinder)
    # reload now to refresh ModuleSpec and associated loader
    if reload:
        assert module_to_reload is not None
        importlib.reload(module_to_reload)
    spec: Optional[importlib.machinery.ModuleSpec] = importlib.util.find_spec(python_module)
    assert spec is not None
    assert spec.loader is not None
    assert isinstance(spec.loader, loader.PluginModuleLoader) == prefer_finder

    # verify that import works and imports the correct module
    mod: ModuleType
    if reload:
        assert module_to_reload is not None
        mod = module_to_reload
    else:
        mod = importlib.import_module(python_module)
    assert mod.where == "libs" if prefer_finder else "venv"


def test_code_loader_prefer_finder(tmpdir: py.path.local, deactive_venv) -> None:
    """
    Verify that the agent code loader prefers its loaded code over code in the Python venv.
    """
    loader.PluginModuleFinder.reset()
    assert not isinstance(sys.meta_path[0], loader.PluginModuleFinder)
    loader.CodeLoader(code_dir=str(tmpdir))
    # it suffices to verify that the module finder is first in the meta path:
    # `test_plugin_module_finder` verifies the actual loader behavior
    assert isinstance(sys.meta_path[0], loader.PluginModuleFinder)


def test_venv_path(tmpdir: py.path.local, projects_dir: str, deactive_venv):
    original_project_dir: str = os.path.join(projects_dir, "plugins_project")
    project_dir = os.path.join(tmpdir, "plugins_project")
    shutil.copytree(original_project_dir, project_dir)

    def load_project(venv_path: str) -> None:
        project: Project = Project(project_dir, venv_path=venv_path)
        Project.set(project)
        # don't load full project, only AST so we don't have to deal with module finder cleanup
        project.install_modules()

    # Use non-default venv dir
    non_default_venv_dir = os.path.join(project_dir, "non-default-venv-dir")
    assert not os.path.exists(non_default_venv_dir)
    load_project(venv_path=non_default_venv_dir)
    assert os.path.exists(non_default_venv_dir)
    shutil.rmtree(non_default_venv_dir)

    # venv_path points to symlink
    os.mkdir(non_default_venv_dir)
    symlink_dir = os.path.join(project_dir, "symlink-dir")
    os.symlink(non_default_venv_dir, symlink_dir)
    for p in [non_default_venv_dir, symlink_dir]:
        assert os.path.exists(p)
        assert not os.path.exists(os.path.join(p, "bin", "python"))
    assert not os.path.islink(non_default_venv_dir)
    assert os.path.islink(symlink_dir)
    load_project(venv_path=symlink_dir)
    for p in [non_default_venv_dir, symlink_dir]:
        assert os.path.exists(os.path.join(p, "bin", "python"))


def test_module_loader(module_path: str, capsys, modules_dir: str):
    """
    Verify that the loader.PluginModuleFinder and loader.PluginModuleLoader load modules correctly.
    """
    origin_mod_dir = os.path.join(modules_dir, "submodule")
    mod_dir = os.path.join(module_path, os.path.basename(origin_mod_dir))
    shutil.copytree(origin_mod_dir, mod_dir)

    capsys.readouterr()  # Clear buffers

    from inmanta_plugins.submodule import test

    assert test() == "test"
    (stdout, stderr) = capsys.readouterr()
    assert stdout.count("#loading inmanta_plugins.submodule#") == 1
    assert stdout.count("#loading inmanta_plugins.submodule.submod#") == 0
    assert stdout.count("#loading inmanta_plugins.submodule.pkg#") == 0
    assert stdout.count("#loading inmanta_plugins.submodule.pkg.submod2#") == 0

    from inmanta_plugins.submodule.submod import test_submod

    assert test_submod() == "test_submod"
    (stdout, stderr) = capsys.readouterr()
    assert stdout.count("#loading inmanta_plugins.submodule#") == 0
    assert stdout.count("#loading inmanta_plugins.submodule.submod#") == 1
    assert stdout.count("#loading inmanta_plugins.submodule.pkg#") == 0
    assert stdout.count("#loading inmanta_plugins.submodule.pkg.submod2#") == 0

    from inmanta_plugins.submodule.pkg import test_pkg

    assert test_pkg() == "test_pkg -- test_submod2"
    (stdout, stderr) = capsys.readouterr()
    assert stdout.count("#loading inmanta_plugins.submodule#") == 0
    assert stdout.count("#loading inmanta_plugins.submodule.submod#") == 0
    assert stdout.count("#loading inmanta_plugins.submodule.pkg#") == 1
    assert stdout.count("#loading inmanta_plugins.submodule.pkg.submod2#") == 1

    with pytest.raises(ImportError):
        from inmanta_plugins.tests import doesnotexist  # NOQA


def test_module_unload(module_path: str, modules_dir: str) -> None:
    """
    Verify that the unload_inmanta_plugins function correctly unloads modules.
    """
    for mod in ["submodule", "elaboratev1module"]:
        origin_mod_dir = os.path.join(modules_dir, mod)
        mod_dir = os.path.join(module_path, os.path.basename(origin_mod_dir))
        shutil.copytree(origin_mod_dir, mod_dir)

    import inmanta_plugins.elaboratev1module  # noqa: F401
    import inmanta_plugins.submodule.submod  # noqa: F401

    assert "inmanta_plugins" in sys.modules
    assert "inmanta_plugins.elaboratev1module" in sys.modules
    assert "inmanta_plugins.submodule" in sys.modules
    assert "inmanta_plugins.submodule.submod" in sys.modules

    loader.unload_inmanta_plugins("submodule")

    assert "inmanta_plugins" in sys.modules
    assert "inmanta_plugins.elaboratev1module" in sys.modules

    assert "inmanta_plugins.submodule" not in sys.modules
    assert "inmanta_plugins.submodule.submod" not in sys.modules

    # make sure that it does not fail on a module with no plugins
    loader.unload_inmanta_plugins("doesnotexist")

    assert "inmanta_plugins" in sys.modules
    assert "inmanta_plugins.elaboratev1module" in sys.modules

    loader.unload_inmanta_plugins()

    assert "inmanta_plugins" not in sys.modules
    assert "inmanta_plugins.elaboratev1module" not in sys.modules


def test_plugin_loading_on_project_load(tmpdir, capsys, deactive_venv):
    """
    Load all plugins via the Project.load() method call and verify that no
    module is loaded twice when an import statement is used.
    """
    main_cf = tmpdir.join("main.cf")
    main_cf.write("import submodule")

    project_yml = tmpdir.join("project.yml")
    project_yml.write(
        """
name: test
modulepath: libs
downloadpath: libs
repo: https://github.com/inmanta/inmanta.git
install_mode: master
    """
    )

    tmpdir.mkdir("libs")
    origin_mod_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data", "modules", "submodule")
    mod_dir = tmpdir.join("libs", os.path.basename(origin_mod_dir))
    shutil.copytree(origin_mod_dir, mod_dir)

    project = Project(tmpdir, autostd=False, venv_path=os.path.join(tmpdir, ".env"))
    Project.set(project)
    project.load()

    (stdout, stderr) = capsys.readouterr()
    assert stdout.count("#loading inmanta_plugins.submodule#") == 1
    assert stdout.count("#loading inmanta_plugins.submodule.submod#") == 1
    assert stdout.count("#loading inmanta_plugins.submodule.pkg#") == 1
    assert stdout.count("#loading inmanta_plugins.submodule.pkg.submod2#") == 1

    from inmanta_plugins.submodule import test

    assert test() == "test"
    (stdout, stderr) = capsys.readouterr()
    assert "#loading" not in stdout

    from inmanta_plugins.submodule.submod import test_submod

    assert test_submod() == "test_submod"
    (stdout, stderr) = capsys.readouterr()
    assert "#loading" not in stdout

    from inmanta_plugins.submodule.pkg import test_pkg

    assert test_pkg() == "test_pkg -- test_submod2"
    (stdout, stderr) = capsys.readouterr()
    assert "#loading" not in stdout


def test_plugin_loading_old_format(tmpdir, capsys):
    """
    Ensure the code loader ignores code formatted in the old on disk format (pre Inmanta 2020.4).
    (See issue: #2162)
    """
    # Create directory structure code dir
    code_dir = tmpdir
    modules_dir = tmpdir.join(loader.MODULE_DIR)
    modules_dir.mkdir()

    # Create source files using pre Inmanta 2020.4 format
    old_format_source_file = modules_dir.join("inmanta_plugins.old_format.py")
    old_format_source_file.write("")

    # Assert code using the pre inmanta 2020.4 format is ignored
    loader.CodeLoader(code_dir)
    with pytest.raises(ImportError):
        import inmanta_plugins.old_format  # NOQA

    # Add newly formatted code next the pre Inmanta 2020.4 format
    new_format_mod_dir = modules_dir.join("new_format")
    new_format_mod_dir.mkdir()
    new_format_plugins_dir = new_format_mod_dir.join("plugins")
    new_format_plugins_dir.mkdir()
    new_format_source_file = new_format_plugins_dir.join("__init__.py")
    new_format_source_file.write(
        """
def test():
    return 10
    """
    )

    # Assert newly formatted code is loaded and code using the pre inmanta 2020.4 format is ignored
    loader.CodeLoader(code_dir)
    import inmanta_plugins.new_format as mod  # NOQA

    assert mod.test() == 10
    with pytest.raises(ImportError):
        import inmanta_plugins.old_format  # NOQA



"""
Copyright 2024 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import abc
import dataclasses
import datetime
import logging
import traceback
import uuid
from collections.abc import Collection
from dataclasses import dataclass

from inmanta import data, resources
from inmanta.agent import executor
from inmanta.agent.executor import DeployReport, FailedInmantaModules
from inmanta.data.model import AttributeStateChange
from inmanta.deploy import scheduler, state
from inmanta.types import ResourceIdStr, ResourceType, ResourceVersionIdStr
from inmanta.vendor import pyformance

LOGGER = logging.getLogger(__name__)


def logger_for_agent(agent: str) -> logging.Logger:
    return logging.getLogger("agent").getChild(agent)


# must remain frozen because it's used as key/identity for deploy intent
@dataclass(frozen=True, kw_only=True)
class Task(abc.ABC):
    """
    Resource action task. Represents the execution of a specific resource action for a given resource.

    Closely coupled with deploy.scheduler.TaskManager interface. Concrete implementations must respect its contract.
    """

    resource: ResourceIdStr

    id: resources.Id = dataclasses.field(init=False, compare=False, hash=False)

    def __post_init__(self) -> None:
        # use object.__setattr__ because this is a frozen dataclass, see dataclasses docs
        object.__setattr__(self, "id", resources.Id.parse_id(self.resource))

    @abc.abstractmethod
    async def execute(self, task_manager: "scheduler.TaskManager", agent: str, reason: str | None = None) -> None:
        pass

    def delete_with_resource(self) -> bool:
        return True

    def get_executor_resource_details(self, version: int, resource_intent: "state.ResourceIntent") -> executor.ResourceDetails:
        return executor.ResourceDetails(
            id=self.resource,
            version=version,
            attributes=resource_intent.attributes,
        )

    async def get_executor(
        self,
        *,
        task_manager: "scheduler.TaskManager",
        agent_spec: tuple[str, Collection[ResourceType]],
        resource_type: ResourceType,
        version: int,
    ) -> executor.Executor:
        """
        Helper method to produce the executor

        :param task_manager: A reference to the task manager instance.
        :param agent_spec: agent name and all resource types that live on it.
        :param resource_type: The resource type that we specifically care about for this resource action.
            Should also be in the agent's resource types.
        :param version: The version of the code to load on the executor.
        """
        agent_name, all_types_for_agent = agent_spec

        if not all_types_for_agent:
            raise ValueError(
                f"{self.__class__.__name__}.get_executor() expects at least one resource type in the agent spec parameter"
            )

        code, invalid_resources = await task_manager.code_manager.get_code(
            environment=task_manager.environment,
            version=version,
            resource_types=all_types_for_agent,
        )

        # Bail out if this failed
        if resource_type in invalid_resources:
            raise invalid_resources[resource_type]

        # Get executor
        my_executor: executor.Executor = await task_manager.executor_manager.get_executor(
            agent_name=agent_name, agent_uri="NO_URI", code=code
        )

        # Bail out if some code failed to load:
        if my_executor.failed_modules:
            raise ModuleLoadingException(
                agent_name=agent_name,
                failed_modules=my_executor.failed_modules,
            )

        return my_executor


class PoisonPill(Task):
    """
    Task to signal queue shutdown

    It is used to make sure all workers wake up to observe that they have been closed.
    It functions mostly as a no-op
    """

    async def execute(self, task_manager: "scheduler.TaskManager", agent: str, reason: str | None = None) -> None:
        pass


class Deploy(Task):
    async def execute(self, task_manager: "scheduler.TaskManager", agent: str, reason: str | None = None) -> None:
        with pyformance.timer("internal.deploy").time():
            # Make id's
            gid = uuid.uuid4()
            action_id = uuid.uuid4()  # can this be gid ?

            # First do scheduler book keeping to establish what to do
            try:
                deploy_intent = await task_manager.deploy_start(action_id, self.resource)
            except Exception:
                # Unrecoverable, can't reach DB
                LOGGER.error(
                    "Failed to report the start of the deployment to the server for %s",
                    self.resource,
                    exc_info=True,
                )
                return

            if deploy_intent is None:
                # Stale resource, can simply be dropped.
                return

            # From this point on, we HAVE to call deploy_done to make sure we are not stuck in deploying
            # We collect state here to report back in the finally block.
            # This try-finally block ensures we report at the end of the task.
            deploy_report: DeployReport
            module_loading_warning: data.LogLine | None = None

            try:
                # Dependencies are always set when calling deploy_start
                assert deploy_intent.dependencies is not None
                # Resolve to executor form
                version: int = deploy_intent.model_version
                resource_intent: "state.ResourceIntent" = deploy_intent.intent
                executor_resource_details: executor.ResourceDetails = self.get_executor_resource_details(
                    version, resource_intent
                )

                # Get executor
                try:
                    # FIXME: code loading interface is not nice like this,
                    #   - we may want to track modules per agent, instead of types
                    #   - we may also want to track the module version vs the model version
                    #       as it avoid the problem of fast chanfing model versions

                    my_executor: executor.Executor = await self.get_executor(
                        task_manager=task_manager,
                        agent_spec=(agent, deploy_intent.all_types_for_agent),
                        resource_type=executor_resource_details.id.entity_type,
                        version=version,
                    )
                except ModuleLoadingException as e:
                    e.log_resource_action_to_scheduler_log(
                        agent=agent, rid=executor_resource_details.rvid, include_exception_info=True
                    )
                    log_line_for_web_console = e.create_log_line_for_failed_modules(level=logging.ERROR, verbose_message=False)
                    deploy_report = DeployReport.undeployable(
                        executor_resource_details.rvid, action_id, log_line_for_web_console
                    )
                    return

                except Exception as e:
                    log_line = data.LogLine.log(
                        logging.ERROR,
                        "All resources of type `%(res_type)s` failed to install handler code "
                        "dependencies: `%(error)s`\n%(traceback)s",
                        res_type=executor_resource_details.id.entity_type,
                        error=str(e),
                        traceback="".join(traceback.format_tb(e.__traceback__)),
                    )

                    # Not attached to ctx, needs to be flushed to logger explicitly
                    log_line.write_to_logger_for_resource(agent, executor_resource_details.rvid, exc_info=True)
                    deploy_report = DeployReport.undeployable(executor_resource_details.rvid, action_id, log_line)

                    return

                if my_executor.failed_modules:

                    # We only create this exception to use its LogLine builder.
                    # We don't raise it since the executor was successfully created for this resource
                    # but we want to log a warning  because not all handler code was successfully loaded.
                    exception = ModuleLoadingException(agent_name=agent, failed_modules=my_executor.failed_modules)
                    exception.log_resource_action_to_scheduler_log(
                        agent=agent, rid=executor_resource_details.rvid, include_exception_info=False
                    )
                    module_loading_warning = exception.create_log_line_for_failed_modules(
                        level=logging.WARNING, verbose_message=False
                    )

                assert reason is not None  # Should always be set for deploy
                # Deploy
                try:
                    deploy_report = await my_executor.execute(
                        action_id, gid, executor_resource_details, reason, deploy_intent.dependencies
                    )

                except Exception as e:
                    # This should not happen

                    # We log both to scheduler log as well as the DB and the resource_action_log
                    # FIXME: can be logging be unified without losing the ability to have this warning prior to writing to DB?
                    # Such that we can have it if the DB is not there
                    LOGGER.error("Failure during executor execution for resource %s", self.resource, exc_info=True)
                    log_line = data.LogLine.log(
                        logging.ERROR,
                        "Failure during executor execution for resource %(res)s",
                        res=self.resource,
                        error=str(e),
                        traceback="".join(traceback.format_tb(e.__traceback__)),
                    )
                    # Not attached to ctx, needs to be flushed to logger explicitly
                    log_line.write_to_logger_for_resource(agent, executor_resource_details.rvid, exc_info=True)
                    deploy_report = DeployReport.undeployable(executor_resource_details.rvid, action_id, log_line)

            finally:
                # We signaled start, so we signal end
                try:
                    if module_loading_warning:
                        deploy_report.messages.append(module_loading_warning)
                    await task_manager.deploy_done(deploy_intent, deploy_report)
                except Exception:
                    LOGGER.error(
                        "Failed to report the end of the deployment to the server for %s",
                        resource_intent.resource_id,
                        exc_info=True,
                    )


@dataclass(frozen=True, kw_only=True)
class DryRun(Task):
    version: int
    resource_intent: state.ResourceIntent
    dry_run_id: uuid.UUID

    def delete_with_resource(self) -> bool:
        return False

    async def execute(self, task_manager: "scheduler.TaskManager", agent: str, reason: str | None = None) -> None:
        executor_resource_details: executor.ResourceDetails = self.get_executor_resource_details(
            self.version, self.resource_intent
        )
        # Just in case we reach the general exception
        started = datetime.datetime.now().astimezone()
        try:
            my_executor: executor.Executor = await self.get_executor(
                task_manager=task_manager,
                agent_spec=(agent, [executor_resource_details.id.entity_type]),
                resource_type=executor_resource_details.id.entity_type,
                version=self.version,
            )
        except Exception:
            logger_for_agent(agent).error(
                "Skipping dryrun for resource %s because due to an error in constructing the executor",
                executor_resource_details.rvid,
                exc_info=True,
            )
            dryrun_report = executor.DryrunReport(
                rvid=executor_resource_details.rvid,
                dryrun_id=self.dry_run_id,
                changes={
                    "handler": AttributeStateChange(
                        current="FAILED", desired="Unable to construct an executor for this resource"
                    )
                },
                started=started,
                finished=datetime.datetime.now().astimezone(),
                messages=[],
            )
        else:
            try:
                dryrun_report = await my_executor.dry_run(executor_resource_details, self.dry_run_id)
            except Exception:
                logger_for_agent(agent).error(
                    "Skipping dryrun for resource %s because it is in undeployable state",
                    executor_resource_details.rvid,
                    exc_info=True,
                )
                dryrun_report = executor.DryrunReport(
                    rvid=executor_resource_details.rvid,
                    dryrun_id=self.dry_run_id,
                    changes={"handler": AttributeStateChange(current="FAILED", desired="Resource is in an undeployable state")},
                    started=started,
                    finished=datetime.datetime.now().astimezone(),
                    messages=[],
                )
        await task_manager.dryrun_done(dryrun_report)


class RefreshFact(Task):

    async def execute(self, task_manager: "scheduler.TaskManager", agent: str, reason: str | None = None) -> None:
        version: int
        version_intent = await task_manager.get_resource_version_intent(self.resource)
        if version_intent is None:
            # Stale resource, can simply be dropped.
            return
        # FIXME, should not need resource intent, only id, see related FIXME on executor side
        version = version_intent.model_version
        resource_intent = version_intent.intent

        executor_resource_details: executor.ResourceDetails = self.get_executor_resource_details(version, resource_intent)
        try:
            my_executor = await self.get_executor(
                task_manager=task_manager,
                agent_spec=(agent, version_intent.all_types_for_agent),
                resource_type=self.id.entity_type,
                version=version,
            )
        except Exception:
            logger_for_agent(agent).warning(
                "Cannot retrieve fact for %s because resource is undeployable or code could not be loaded",
                executor_resource_details.rvid,
            )
            return

        get_fact_report = await my_executor.get_facts(executor_resource_details)
        await task_manager.fact_refresh_done(get_fact_report)


class ModuleLoadingException(Exception):
    """
    This exception is raised when some Inmanta modules couldn't be loaded on a given agent.
    """

    def __init__(self, agent_name: str, failed_modules: FailedInmantaModules) -> None:
        """
        :param agent_name: Name of the agent for which module loading was unsuccessful
        :param failed_modules: Data for all module loading errors as a nested map of
            inmanta module name -> python module name -> Exception.
        """
        self.failed_modules = failed_modules
        self.agent_name = agent_name

    def _format_module_loading_errors(self) -> str:
        """
        Helper method to display module loading failures.
        """
        formatted_module_loading_errors = ""
        N_FAILURES = sum(len(v) for v in self.failed_modules.values())
        failure_index = 1

        for _, failed_modules_data in self.failed_modules.items():
            for python_module, exception in failed_modules_data.items():
                formatted_module_loading_errors += f"Error {failure_index}/{N_FAILURES}:\n"
                formatted_module_loading_errors += f"In module {python_module}:\n"
                formatted_module_loading_errors += str(exception)
                failure_index += 1

        return formatted_module_loading_errors

    def create_log_line_for_failed_modules(self, level: int = logging.ERROR, *, verbose_message: bool = False) -> data.LogLine:
        """
        Helper method to convert this Exception into a LogLine.

        :param level: The log level for the resulting LogLine
        :param verbose_message: Whether to include the full formatted error output in the LogLine message.
            When displayed on the webconsole, the full formatted error output will be displayed in its own section
            regardless of this flag's value.

        """
        message = "Agent %s failed to load the following modules: %s." % (
            self.agent_name,
            ", ".join(self.failed_modules.keys()),
        )

        formatted_module_loading_errors = self._format_module_loading_errors()

        if verbose_message:
            message += f"\n{formatted_module_loading_errors}"
        return data.LogLine.log(
            level=level,
            msg=message,
            timestamp=None,
            errors=formatted_module_loading_errors,
        )

    def log_resource_action_to_scheduler_log(
        self, agent: str, rid: ResourceVersionIdStr, *, include_exception_info: bool
    ) -> None:
        """
        Helper method to log module loading failures to the scheduler's resource action log.

        This method does not write anything to the 'resourceaction' table, which is what is ultimately displayed in the
        'Logs' tab of the 'Resource Details' page in the web console. Therefore, the caller is responsible
        for making sure that the scheduler's resource action log and the web console logs remain somewhat in sync.

        """
        log_line = self.create_log_line_for_failed_modules(verbose_message=True)
        log_line.write_to_logger_for_resource(agent=agent, resource_version_string=rid, exc_info=include_exception_info)



"""
Copyright 2014 Omer Gertel
Copyright 2025 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com

This code was originally developed by Omer Gertel, as a python port of the core portion of a
[Java Metrics library by Coda Hale](http://metrics.dropwizard.io/)

It was vendored into the inmanta source tree as the original was no longer maintained.
"""

import math
from collections.abc import Iterable


class Snapshot(object):
    """
    This class is used by the histogram meter
    """

    MEDIAN = 0.5
    P75_Q = 0.75
    P95_Q = 0.95
    P99_Q = 0.99
    P999_Q = 0.999

    def __init__(self, values: Iterable[float]) -> None:
        super(Snapshot, self).__init__()
        self.values = sorted(values)

    def get_size(self) -> int:
        "get current size"
        return len(self.values)

    def get_sum(self) -> float:
        "get current sum"
        return float(sum(self.values))

    def get_max(self) -> float:
        "get current maximum value"
        if not self.values:
            return 0
        return self.values[-1]

    def get_min(self) -> float:
        "get current minimum value"
        if not self.values:
            return 0
        return self.values[0]

    def get_mean(self) -> float:
        "get current mean value"
        if not self.values:
            return 0
        return float(sum(self.values)) / self.get_size()

    def get_stddev(self) -> float:
        "get current standard deviation"
        if not self.values:
            return 0
        return math.sqrt(self.get_var())

    def get_var(self) -> float:
        "get current variance"
        if not self.values or self.get_size() == 1:
            return 0
        mean = self.get_mean()
        square_differences = [(mean - value) ** 2 for value in self.values]
        return sum(square_differences) / (self.get_size() - 1)

    def get_median(self) -> float:
        "get current median"
        return self.get_percentile(Snapshot.MEDIAN)

    def get_75th_percentile(self) -> float:
        "get current 75th percentile"
        return self.get_percentile(Snapshot.P75_Q)

    def get_95th_percentile(self) -> float:
        "get current 95th percentile"
        return self.get_percentile(Snapshot.P95_Q)

    def get_99th_percentile(self) -> float:
        "get current 99th percentile"
        return self.get_percentile(Snapshot.P99_Q)

    def get_999th_percentile(self) -> float:
        "get current 999th percentile"
        return self.get_percentile(Snapshot.P999_Q)

    def get_percentile(self, percentile: float) -> float:
        """
        get custom percentile

        :param percentile: float value between 0 and 1
        """
        if percentile < 0 or percentile > 1:
            raise ValueError("{0} is not in [0..1]".format(percentile))
        length = len(self.values)
        if length == 0:
            return 0
        pos = percentile * (length + 1)
        if pos < 1:
            return self.values[0]
        if pos >= length:
            return self.values[-1]
        lower = self.values[int(pos) - 1]
        upper = self.values[int(pos)]
        return lower + (pos - int(pos)) * (upper - lower)



"""
Copyright 2017 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

from abc import abstractmethod
from collections.abc import Hashable, Sequence, Set
from typing import TYPE_CHECKING, Deque, Generic, List, Literal, NewType, Optional, TypeVar, Union, cast

import inmanta.ast
import inmanta.ast.attribute  # noqa: F401 (pyflakes does not recognize partially qualified access ast.attribute)
from inmanta import ast, references
from inmanta.ast import (
    AttributeException,
    CompilerException,
    DoubleSetException,
    Locatable,
    Location,
    ModifiedAfterFreezeException,
    Namespace,
    NotFoundException,
    OptionalValueException,
    RuntimeException,
)
from inmanta.ast.type import Type
from inmanta.execute import dataflow
from inmanta.execute.dataflow import DataflowGraph
from inmanta.execute.util import NoneValue, Unknown

if TYPE_CHECKING:
    from inmanta.ast.blocks import BasicBlock
    from inmanta.ast.entity import Entity, Implementation
    from inmanta.ast.statements import RawResumer, RequiresEmitStatement, Resumer, Statement
    from inmanta.compiler import Compiler
    from inmanta.execute.scheduler import PrioritisedDelayedResultVariableQueue


T = TypeVar("T")
T_co = TypeVar("T_co", covariant=True)
T_contra = TypeVar("T_contra", contravariant=True)


class ResultCollector(Generic[T_contra]):
    """
    Helper interface for gradual execution. Should be attached as a listener to a ResultVariable, which will then call
    receive_result whenever it receives a new value.
    """

    __slots__ = ()

    def pure_gradual(self) -> bool:
        """
        Returns true iff this result collector represents pure gradual execution, i.e. all progress comes from new values and
        no progress is expected as a result of the variable being frozen.

        This is a static property of a result collector, i.e. this method will always have the same result at any point within
        the lifetime of the instance.
        """
        return True

    def receive_result(self, value: T_contra, location: Location) -> bool:
        """
        Receive a single value for gradual execution. Called once for each value that is part of the result. May be `null` or
        Unknown. Does not distinguish between Unknown as an element of a list and Unknown as a top-level value.

        :return: Whether this collector is complete, i.e. it does not need to receive any further results and its associated
            waiter will no longer cause progress. Once this is signalled, this instance should get no further results.
        """
        raise NotImplementedError()

    def receive_result_flatten(self, value: Union[T_contra, Sequence[T_contra]], location: Location) -> bool:
        """
        Receive one or more values for gradual execution. When a list is passed, its values will be passed to receive_result
        one by one. Otherwise the value itself is passed unmodified.

        :return: Whether this collector is complete, i.e. it does not need to receive any further results and its associated
            waiter will no longer cause progress. Once this is signalled, this instance should get no further results.
        """
        for subvalue in value if isinstance(value, list) else [value]:
            done: bool = self.receive_result(subvalue, location)
            if done:
                return True
        return False


class IPromise:
    """
    A promise to the owner to provide a value or progression towards a value in some way, either directly or indirectly.
    To provide strict provider tracking, overpromising is allowed in case of uncertainty as to whether progression towards
    a value will be made, as long as progression towards certainty is made.
    """

    __slots__ = ()


class ISetPromise(IPromise, Generic[T]):
    """
    A promise to the owner to set a value.
    """

    __slots__ = ()

    @abstractmethod
    def set_value(self, value: T, location: Location) -> None:
        """
        Fulfills this promise by setting the owner's value and notifying the owner of the promise's completion.
        """


class ProgressionPromise(IPromise):
    """
    A promise from a provider to the owner to progress towards setting a value, for example by emitting additional statements.
    """

    __slots__ = ("provider", "owner")

    def __init__(self, owner: "ResultVariable[T]", provider: "Statement") -> None:
        self.owner: ResultVariable[T] = owner
        self.provider: Statement = provider

    def fulfill(self) -> None:
        """
        Fulfills this promise by notifying the owner. No further progression is expected. Idempotent.
        """
        self.owner.fulfill(self)


class VariableABC(Generic[T_co]):
    """
    Abstract base class for variables that get passed around the AST nodes' methods via waiters.
    """

    __slots__ = ()

    def is_ready(self) -> bool:
        """
        Returns true iff this variable does not expect any more values.
        """
        raise NotImplementedError()

    def get_value(self) -> T_co:
        """
        Returns the value object for this variable

        :raises OptionalValueException: This is an optional variable that has not received a value (or explicit `null`).
        """
        raise NotImplementedError()

    def listener(self, resultcollector: ResultCollector[T_co], location: Location) -> bool:
        """
        Add a listener to report new values to. If the variable already has a value, this is reported immediately. Explicit
        assignments of `null` will not be reported.

        Each listener is expected to register one associated waiter to track completeness. The progress potential implementation
        is based on this invariant.

        :param resultcollector: The collector for the values of this variable.
        :param location: The location associated with this listener.
        :return: True iff this variable will report values to the listener. Simple variables may not implement the listen
            capability and will return False. A variable will always return the same value regardless of the parameters.
        """
        return False

    def waitfor(self, waiter: "Waiter") -> None:
        """
        Informs this variable that a waiter waits on its value. Once the variable receives a value, it should inform the waiter.
        """
        raise NotImplementedError()

    def get_progression_promise(self, provider: "Statement") -> Optional[ProgressionPromise]:
        """
        Acquires a promise to progress this variable without necessarily setting a value. It is allowed to acquire a progression
        promise greedily (overpromise) when a provider is likely to produce progress. The promise should then be fulfilled as
        soon as it is known that no further progress will be made.

        Returns None if this variable does not track progression promises.

        e.g. a progression promise could be acquired by a conditional statement that might emit a new assignment statement for
        this variable. As soon as the condition is evaluated this promise should be fulfilled.

        This overpromising semantics allows for more strict promise tracking, providing more certainty on variable completeness
        at the cost of disallowing circular logic.
        """
        return None


class WrappedValueVariable(VariableABC[T]):
    """
    Variable that holds a single value that is known at construction. Used to wrap values where a VariableABC is expected.
    """

    __slots__ = ("value",)

    def __init__(self, value: T) -> None:
        self.value: T = value

    def is_ready(self) -> bool:
        return True

    def get_value(self) -> T:
        return self.value

    def listener(self, resultcollector: ResultCollector[T], location: Location) -> Literal[True]:
        resultcollector.receive_result(self.value, location)
        return True

    def waitfor(self, waiter: "Waiter") -> None:
        waiter.ready(self)


class ResultVariable(VariableABC[T], ResultCollector[T], ISetPromise[T]):
    """
    A ResultVariable is like a future
     - it has a list of waiters
     - when a value is set, the waiters are notified,
        they decrease their wait count and
        queue themselves when their wait count becomes 0

    If a type is set on a result variable, setting a value of another type will produce an exception.

    In order to assist heuristic evaluation, result variables keep track of any statement that will assign a value to it
    """

    location: Location

    __slots__ = ("location", "provider", "waiters", "value", "hasValue", "type", "_node")

    def __init__(self, value: Optional[T] = None) -> None:
        self.waiters: "List[Waiter]" = []
        self.value: Optional[T] = value
        self.hasValue: bool = False
        self.type: Optional[Type] = None
        self._node: Optional[dataflow.AssignableNodeReference] = None

    def set_type(self, mytype: Type) -> None:
        self.type = mytype

    def get_promise(self, provider: "Statement") -> ISetPromise[T]:
        """
        Acquire a promise to set a value for this variable. To fulfill the promise and set the promised value for this
        variable, set the value on the promise object.
        """
        return self

    def fulfill(self, promise: IPromise) -> None:
        """
        Considers the given promise fulfilled. Idempotent. Should only be called with promises that were handed out by thie
        variable.
        """
        # plain ResultVariable does not track promises -> simply return

    def is_ready(self) -> bool:
        return self.hasValue

    def waitfor(self, waiter: "Waiter") -> None:
        if self.is_ready():
            waiter.ready(self)
        else:
            self.waiters.append(waiter)

    def set_value(self, value: T, location: Location, recur: bool = True) -> None:
        """
        Set the value for this result variable.

        :param recur: If True, recur on the other side of this variable if it is part of a bidirectional relation.
        """
        if self.hasValue:
            if self.value != value:
                raise DoubleSetException(self, None, value, location)
            else:
                return
        if not isinstance(value, Unknown) and self.type is not None:
            self.type.validate(value)
        self.value = value
        self.location = location
        self.hasValue = True
        for waiter in self.waiters:
            waiter.ready(self)

        # prevent memory leaks
        self.waiters = None

    def get_value(self) -> T:
        if not self.hasValue:
            raise inmanta.ast.UnsetException("Value not available", self)

        return self.value

    def can_get(self) -> bool:
        return self.hasValue

    def freeze(self) -> None:
        pass

    def receive_result(self, value: T, location: Location) -> bool:
        return True

    def is_multi(self) -> bool:
        return False

    def set_dataflow_node(self, node: dataflow.AssignableNodeReference) -> None:
        assert self._node is None or self._node == node
        self._node = node
        self._node.set_result_variable(self)

    def get_dataflow_node(self) -> dataflow.AssignableNodeReference:
        assert self._node is not None, "assertion error at %s.get_dataflow_node() in ResultVariable" % self
        return self._node


class ListElementVariable(ResultVariable[T]):
    """
    Variable representing a single value in a gradually executable list. Can be resolved immediately and allows attaching
    listeners who will receive this single value.
    """

    def __init__(self, value: T, location: Location) -> None:
        super().__init__()
        self.set_value(value, location)

    def listener(self, resultcollector: ResultCollector[T], location: Location) -> Literal[True]:
        assert self.value is not None
        resultcollector.receive_result(self.value, location)
        return True


class ResultVariableProxy(VariableABC[T]):
    """
    A proxy for a reading from a ResultVariable that implements the VariableABC interface. Allows for assignment between
    variables without resolving the right hand side at the time of assignment. Supports a single listener to be registered
    at construction. This class will ensure that the connected variable's value(s) is (are) reported to the listener, even if
    the variable itself does not support listeners.
    This class does not support setting values or related operations such as acquiring progression promises.
    """

    __slots__ = ("variable", "_listener", "_waiters", "_notify_listeners")

    def __init__(
        self, variable: Optional[VariableABC[T]] = None, listener: Optional[tuple[ResultCollector[T], Location]] = None
    ) -> None:
        self.variable: Optional[VariableABC[T]] = variable
        self._listener: Optional[tuple[ResultCollector[T], Location]] = listener
        self._waiters: Optional[list["Waiter"]] = []
        # are we responsible for notifying listeners ourselves?
        self._notify_listeners: bool = False

    def connect(self, variable: VariableABC[T]) -> None:
        """
        Connect this proxy to a variable. A proxy can only be connected to a single variable.
        """
        if self.variable is not None and self.variable != variable:
            raise Exception("Trying to connect a variable to a proxy that is already connected to another variable.")
        self.variable = variable
        assert self._waiters is not None  # only set to None after a variable is connected to prevent data leaks
        if self._listener is not None:
            registered_listener: bool = self.variable.listener(*self._listener)
            if registered_listener:
                # clear listener to prevent data leaks
                self._listener = None
            else:
                # variable does not support listeners, we'll have to notify the listener ourselves
                self._notify_listeners = True
        for waiter in self._waiters:
            self.variable.waitfor(waiter)
        # clear waiters to prevent data leaks
        self._waiters = None

    def is_ready(self) -> bool:
        return self.variable is not None and self.variable.is_ready()

    def get_value(self) -> T:
        """
        Returns the value object for this variable
        """
        if self.variable is None:
            raise Exception(
                "Trying to get value for proxy variable that has not been connected yet. Use `waitfor` to wait for a value."
            )
        value: T = self.variable.get_value()
        if self._notify_listeners:
            assert self._listener is not None
            listener, location = self._listener
            # simple case: single value. Multi-value variables implement their own listener functionality
            listener.receive_result_flatten(value, location)
            # clean up: prevent data leaks and ensure listener is only notified once
            self._listener = None
            self._notify_listeners = False
        return value

    def listener(self, resultcollector: ResultCollector[T], location: Location) -> Literal[False]:
        # no need for this right now and implementing it greatly complicates the logic of this class
        # a single listener can be registered at construction time
        return False

    def waitfor(self, waiter: "Waiter") -> None:
        """
        Informs this variable that a waiter waits on its value. Once the variable receives a value, it should inform the waiter.
        """
        if self.variable is None:
            assert self._waiters is not None  # only set to None after a variable is connected to prevent data leaks
            self._waiters.append(waiter)
        else:
            self.variable.waitfor(waiter)


class RelationAttributeVariable:
    """
    Abstract base class for variables associated with a relation attribute.
    """

    __slots__ = ()


class AttributeVariable(ResultVariable["Instance"], RelationAttributeVariable):
    """
    a result variable for a relation with arity 1

    when assigned a value, it will also assign a value to its inverse relation
    """

    __slots__ = ("attribute", "myself")

    def __init__(self, attribute: "ast.attribute.RelationAttribute", instance: "Instance"):
        self.attribute: ast.attribute.RelationAttribute = attribute
        self.myself: "Instance" = instance
        ResultVariable.__init__(self)

    def set_value(self, value: "Instance", location: Location, recur: bool = True) -> None:
        if self.hasValue:
            if self.value != value:
                raise DoubleSetException(self, None, value, location)
            else:
                return
        if not isinstance(value, Unknown) and self.type is not None:
            self.type.validate(value)
        self.value = value
        self.location = location
        self.hasValue = True
        # set counterpart
        if self.attribute.end and recur:
            assert isinstance(value, Instance)
            value.set_attribute(self.attribute.end.name, self.myself, location, False)
        for waiter in self.waiters:
            waiter.ready(self)
        self.waiters = None


class SetPromise(ISetPromise[T]):
    """
    A promise from a provider to the owner to set a value.
    """

    __slots__ = ("provider", "owner")

    def __init__(self, owner: "DelayedResultVariable[T]", provider: "Statement"):
        self.provider: "Optional[Statement]" = provider
        self.owner: DelayedResultVariable[T] = owner

    def set_value(self, value: T, location: Location) -> None:
        self.owner.set_value(value, location, recur=True)
        self.owner.fulfill(self)


class DelayedResultVariable(ResultVariable[T]):
    """
    DelayedResultVariable are ResultVariables of which it is unclear how many results will be set.

    i.e. there may be speculation about when they can be considered complete.

    When the freeze method is called, no more values will be accepted and
    the DelayedResultVariable will behave as a normal ResultVariable.

    When a DelayedResultVariable is definitely full, it freezes itself.

    DelayedResultVariable are queued with the scheduler at the point at which they might be complete.
    The scheduler can decide when to freeze them. A DelayedResultVariable  can be complete when
      - it contains enough elements
      - there are no providers which still have to provide some values (tracked inexactly)
        (a queue variable can be dequeued by the scheduler when a provider is added)
    """

    __slots__ = ("queued", "queues", "promises", "done_promises")

    def __init__(self, queue: "QueueScheduler", value: Optional[T] = None) -> None:
        ResultVariable.__init__(self, value)
        self.promises: Optional[list[IPromise]] = []
        self.done_promises: Optional[set[IPromise]] = set()
        self.queued = False
        self.queues = queue
        if self.can_get():
            self.queue()

    def get_promise(self, provider: "Statement") -> ISetPromise[T]:
        promise: ISetPromise[T] = SetPromise(self, provider)
        if self.promises is not None:
            # only track the promise if this variable has not been frozen yet.
            self.promises.append(promise)
        return promise

    def get_progression_promise(self, provider: "Statement") -> Optional[ProgressionPromise]:
        if self.promises is None:
            return None
        promise: ProgressionPromise = ProgressionPromise(self, provider)
        self.promises.append(promise)
        return promise

    def fulfill(self, promise: IPromise) -> None:
        if self.done_promises is None:
            # already frozen, no need to track promises anymore
            return
        self.done_promises.add(promise)
        if self.can_get():
            self.queue()

    def freeze(self) -> None:
        if self.hasValue:
            return
        self.queued = True
        self.hasValue = True
        for waiter in self.waiters:
            waiter.ready(self)
        # prevent memory leaks
        self.waiters = None
        self.queues = None
        self.promises = None
        self.done_promises = None

    def queue(self) -> None:
        if self.queued:
            return
        self.queued = True
        self.queues.add_possible(self)

    def unqueue(self) -> None:
        self.queued = False

    def get_waiting_providers(self) -> int:
        """How many values are definitely still waiting for"""
        if self.promises is None:
            # already frozen
            return 0
        # todo: optimize?
        assert self.done_promises is not None
        out = len(self.promises) - len(self.done_promises)
        if out < 0:
            raise Exception("SEVERE: COMPILER STATE CORRUPT: provide count negative")
        return out

    def get_progress_potential(self) -> int:
        """
        Returns the number of blocked waiters, meaning any waiters that may progress when this variable is frozen.
        """
        return len(self.waiters)


ListValue = Union["Instance", list["Instance"]]


class BaseListVariable(DelayedResultVariable[ListValue]):
    """
    List variable, but only the part that is independent of an instance
    """

    value: "List[Instance]"

    __slots__ = ("_listeners", "_nb_gradual_waiters")

    def __init__(self, queue: "QueueScheduler") -> None:
        # use dict for easy lookup with reliable ordering
        self._listeners: Optional[dict[ResultCollector["Instance"], None]] = {}
        # Cache count for waiters without progress potential. Meaning waiters associated with either a purely gradual
        # listener or with a listener that indicated it is done.
        self._nb_gradual_waiters: int = 0
        super().__init__(queue, [])

    def _set_value(self, value: ListValue, location: Location, recur: bool = True) -> bool:
        """
        First half of set_value, returns True if second half should run
        """
        if self.hasValue:
            if isinstance(value, list):
                if len(value) == 0:
                    # empty list terminates list addition
                    return False
                for subvalue in value:
                    if subvalue not in self.value:
                        raise ModifiedAfterFreezeException(
                            self,
                            instance=self.myself,
                            attribute=self.attribute,
                            value=value,
                            location=location,
                            reverse=not recur,
                        )
            elif value in self.value:
                return False
            else:
                raise ModifiedAfterFreezeException(
                    self, instance=self.myself, attribute=self.attribute, value=value, location=location, reverse=not recur
                )

        if isinstance(value, list):
            if len(value) == 0:
                # the values of empty lists need no processing,
                # but a set_value from an empty list may fulfill a promise, allowing this object to be queued
                if self.can_get():
                    self.queue()
            else:
                for v in value:
                    self.set_value(v, location, recur)
            return False

        if self.type is not None:
            self.type.validate(value)

        if value in self.value:
            # any set_value may fulfill a promise, allowing this object to be queued
            if self.can_get():
                self.queue()
            return False

        self.value.append(value)
        self._notify_listeners(value, location)

        return True

    def _notify_listeners(self, value: "Instance", location: Location) -> None:
        """
        Notifies all listeners about a new value. Deregisters any listeners that indicate they no longer wish to receive
        results.
        """
        assert self._listeners is not None
        for listener in list(self._listeners.keys()):
            done: bool = listener.receive_result(value, location)
            if done:
                if not listener.pure_gradual():
                    # listener used to have progress potential but not anymore
                    self._nb_gradual_waiters += 1
                # keep memory footprint minimal
                del self._listeners[listener]

    def set_value(self, value: ListValue, location: Location, recur: bool = True) -> None:
        if not self._set_value(value, location, recur):
            return
        if self.can_get():
            self.queue()

    def can_get(self) -> bool:
        return self.get_waiting_providers() == 0

    def receive_result(self, value: ListValue, location: Location) -> bool:
        self.set_value(value, location)
        return False

    def listener(self, resultcollector: ResultCollector["Instance"], location: Location) -> Literal[True]:
        for value in self.value:
            resultcollector.receive_result(value, location)
        if not self.hasValue:
            assert self._listeners is not None
            if resultcollector in self._listeners:
                # may happen in case of a duplicate assignment, e.g. `x.a = [y.a, y.a]`
                # consider the new one to have no progress potential because we don't track it separately
                self._nb_gradual_waiters += 1
                return True
            self._listeners[resultcollector] = None
            if resultcollector.pure_gradual():
                self._nb_gradual_waiters += 1
        return True

    def is_multi(self) -> bool:
        return True

    def get_progress_potential(self) -> int:
        # purely gradual waiters aren't blocked on this variable being frozen
        return len(self.waiters) - self._nb_gradual_waiters

    def freeze(self) -> None:
        super().freeze()
        # prevent memory leaks
        self._listeners = None

    def __str__(self) -> str:
        return "BaseListVariable %s" % (self.value)


# known issue: typed as ResultVariable[ListValue] but is actually ResultVariable[object]
class ListLiteral(BaseListVariable):
    """
    Transient variable to represent a list (of either constants or instances) literal (not a variable).
    Requires all providers to acquire a promise before the first gets fulfilled and in return provides accurate promise
    tracking and freezing. Instances of this class should never require forceful freezing.
    """

    __slots__ = ()

    def fulfill(self, promise: IPromise) -> None:
        """
        Fulfill a promise with 100% accurate promise tracking. Because of this class' invariant that all promises are
        acquired before the first is fulfilled, the list can safely be frozen once all registered promises have been fulfilled.
        """
        super().fulfill(promise)
        # 100% accurate promisse tracking
        if self.get_waiting_providers() == 0:
            self.freeze()

    def __str__(self) -> str:
        return "TempListVariable %s" % (self.value)


class ListVariable(BaseListVariable, RelationAttributeVariable):
    """
    ResultVariable that represents a list of instances associated with a relation attribute.
    """

    value: "List[Instance]"

    __slots__ = ("attribute", "myself")

    def __init__(self, attribute: "ast.attribute.RelationAttribute", instance: "Instance", queue: "QueueScheduler") -> None:
        self.attribute: ast.attribute.RelationAttribute = attribute
        self.myself: "Instance" = instance
        BaseListVariable.__init__(self, queue)

    def set_value(self, value: ListValue, location: Location, recur: bool = True) -> None:
        if isinstance(value, NoneValue):
            if len(self.value) > 0:
                exception: CompilerException = RuntimeException(
                    None,
                    "Trying to set relation attribute `%s` of instance `%s` to null but it has values `%s` assigned already"
                    % (self.attribute.name, self.myself, ", ".join(str(v) for v in self.value)),
                )
                exception.set_location(location)
                raise exception
            self.freeze()
            return
        try:
            if not self._set_value(value, location, recur):
                return
        except ModifiedAfterFreezeException as e:
            if len(self.value) == self.attribute.high:
                new_exception: CompilerException = RuntimeException(
                    None, f"Exceeded relation arity on attribute '{self.attribute.name}' of instance '{self.myself}'"
                )
                new_exception.set_location(location)
                raise new_exception
            raise e
        # set counterpart
        if self.attribute.end is not None and recur:
            value.set_attribute(self.attribute.end.name, self.myself, location, False)

        if self.attribute.high is not None:
            if len(self.value) > self.attribute.high:
                raise RuntimeException(
                    None, "List over full: max nr of items is %d, content is %s" % (self.attribute.high, self.value)
                )

            if self.attribute.high == len(self.value):
                self.freeze()

        if self.can_get():
            self.queue()

    def can_get(self) -> bool:
        return len(self.value) >= self.attribute.low and self.get_waiting_providers() == 0

    def __str__(self) -> str:
        return f"ListVariable {self.myself} {self.attribute} = {self.value}"

    def get_progress_potential(self) -> int:
        # Ensure that relationships with a relation precedence rule cannot end up in the zerowaiters queue
        # of the scheduler. We know the order in which those types can be frozen safely.
        return super().get_progress_potential() + int(self.attribute.has_relation_precedence_rules())


class OptionVariable(DelayedResultVariable["Instance"], RelationAttributeVariable):
    """
    Variable to hold the value for an optional relation (arity [0:1])
    """

    __slots__ = ("attribute", "myself", "location")

    def __init__(self, attribute: "ast.attribute.Attribute", instance: "Instance", queue: "QueueScheduler") -> None:
        self.value = None
        self.attribute: ast.attribute.RelationAttribute = attribute
        self.myself: "Instance" = instance
        self.location = None
        # Only call super after initialization of the above-mentioned attributes
        # because the self.queue() operation in DelayedResultVariable requires
        # self.attribute and self.myself to be set.
        DelayedResultVariable.__init__(self, queue)

    def _get_null_value(self) -> object:
        return None

    def set_value(self, value: object, location: Location, recur: bool = True) -> None:
        assert location is not None
        if self.hasValue:
            if self.value is None:
                if isinstance(value, NoneValue):
                    return
                else:
                    raise ModifiedAfterFreezeException(
                        self, instance=self.myself, attribute=self.attribute, value=value, location=location, reverse=not recur
                    )
            elif self.value != value:
                raise DoubleSetException(self, None, value, location)

        self._validate_value(value)

        if isinstance(value, Instance):
            # set counterpart
            if self.attribute.end is not None and recur:
                value.set_attribute(self.attribute.end.name, self.myself, location, False)

        self.value = value if not isinstance(value, NoneValue) else self._get_null_value()
        self.location = location
        self.freeze()

    def _validate_value(self, value: object) -> None:
        if isinstance(value, Unknown):
            return
        if isinstance(value, NoneValue):
            return
        if self.type is None:
            return
        self.type.validate(value)

    def can_get(self) -> bool:
        return self.get_waiting_providers() == 0

    def get_value(self) -> "Instance":
        result = DelayedResultVariable.get_value(self)
        if result is None:
            raise OptionalValueException(self.myself, self.attribute)
        return result

    def __str__(self) -> str:
        return f"OptionVariable {self.myself} {self.attribute} = {self.value}"

    def get_progress_potential(self) -> int:
        return super().get_progress_potential() + int(self.attribute.has_relation_precedence_rules())


WaiterSet = NewType("WaiterSet", Set["Waiter"])
"""
Set-like object with deterministic iteration order (maintains insert order).

Declared as NewType rather than an implementation of MutableSet for performance reasons (one order of magnitude).
Constructor is responsible for upholding the type's semantics.
"""


class QueueScheduler:
    """
    Object representing the compiler to the AST nodes. It provides access to the queueing mechanism and the type system.

    MUTABLE!
    """

    __slots__ = ("compiler", "runqueue", "waitqueue", "types", "_allwaiters")

    def __init__(
        self,
        compiler: "Compiler",
        runqueue: "Deque[Waiter]",
        waitqueue: "PrioritisedDelayedResultVariableQueue",
        types: dict[str, Type],
    ) -> None:
        self.compiler = compiler
        self.runqueue = runqueue
        self.waitqueue = waitqueue
        self.types = types
        # store internally as dict rather than set to achieve deterministic iteration order
        self._allwaiters: dict[Waiter, None] = {}

    @property
    def allwaiters(self) -> WaiterSet:
        return WaiterSet(self._allwaiters.keys())

    def add_running(self, item: "Waiter") -> None:
        self.runqueue.append(item)

    def add_possible(self, rv: DelayedResultVariable) -> None:
        self.waitqueue.append(rv)

    def get_compiler(self) -> "Compiler":
        return self.compiler

    def get_types(self) -> dict[str, Type]:
        return self.types

    def add_to_all(self, item: "Waiter") -> None:
        self._allwaiters[item] = None

    def remove_from_all(self, item: "Waiter") -> None:
        del self._allwaiters[item]


class Waiter:
    """
    Waiters represent an executable unit, that can be executed the result variables they depend on have their values.
    """

    __slots__ = ("waitcount", "queue", "done", "requires")

    requires: dict[object, VariableABC]

    def __init__(self, queue: QueueScheduler):
        self.waitcount = 1
        self.queue = queue
        self.queue.add_to_all(self)
        self.done = False

    def requeue_with_additional_requires(self, key: Hashable, waitable: ResultVariable) -> None:
        """
        Re-queue with an additional requirement
        """
        self.requires[key] = waitable
        self.waitfor(waitable)

    def waitfor(self, waitable: VariableABC[object]) -> None:
        self.waitcount = self.waitcount + 1
        waitable.waitfor(self)

    def ready(self, other: Union[ResultVariable, "Waiter"]) -> None:
        self.waitcount = self.waitcount - 1
        if self.waitcount == 0:
            self.queue.add_running(self)
        if self.waitcount < 0:
            raise Exception("SEVERE: COMPILER STATE CORRUPT: waitcount negative")

    def execute(self) -> None:
        pass


class ExecutionUnit(Waiter):
    """
    Basic assign statement:
     - Wait for a dict of requirements
     - Call the execute method on the expression, with a map of the resulting values
     - Assign the resulting value to the result variable

     @param provides: Whether to register this XU as provider to the result variable
    """

    __slots__ = ("result", "expression", "resolver", "queue_scheduler", "owner")

    def __init__(
        self,
        queue_scheduler: QueueScheduler,
        resolver: "Resolver",
        result: ResultVariable[object],
        requires: dict[object, VariableABC],
        expression: "RequiresEmitStatement",
        owner: "Optional[Statement]" = None,
    ):
        Waiter.__init__(self, queue_scheduler)
        self.result: ISetPromise[object] = result.get_promise(expression)
        self.requires = requires
        self.expression = expression
        self.resolver = resolver
        self.queue_scheduler = queue_scheduler
        for r in requires.values():
            self.waitfor(r)
        self.ready(self)
        if owner is not None:
            self.owner = owner
        else:
            self.owner = expression

    def _unsafe_execute(self) -> None:
        requires = {k: v.get_value() for (k, v) in self.requires.items()}
        value = self.expression.execute(requires, self.resolver, self.queue_scheduler)
        self.result.set_value(value, self.expression.location)
        self.done = True

    def execute(self) -> None:
        try:
            self._unsafe_execute()
        except RuntimeException as e:
            e.set_statement(self.owner, False)
            raise e

    def __repr__(self) -> str:
        return repr(self.expression)


class HangUnit(Waiter):
    """
    Wait for a dict of requirements, call the resume method on the resumer, with a map of the resulting values
    """

    __slots__ = ("resolver", "resumer", "target")

    def __init__(
        self,
        queue_scheduler: QueueScheduler,
        resolver: "Resolver",
        requires: dict[object, VariableABC],
        target: Optional[ResultVariable],
        resumer: "Resumer",
    ) -> None:
        Waiter.__init__(self, queue_scheduler)
        self.resolver = resolver
        self.requires = requires
        self.resumer = resumer
        self.target = target
        for r in requires.values():
            self.waitfor(r)
        self.ready(self)

    def execute(self) -> None:
        try:
            self.resumer.resume({k: v.get_value() for (k, v) in self.requires.items()}, self.resolver, self.queue, self.target)
        except RuntimeException as e:
            e.set_statement(self.resumer)
            raise e
        self.done = True


class RawUnit(Waiter):
    """
    Wait for a map of requirements, call the resume method on the resumer,
    but with a map of ResultVariables instead of their values
    """

    __slots__ = ("resolver", "resumer", "override_exception_location")

    def __init__(
        self,
        queue_scheduler: QueueScheduler,
        resolver: "Resolver",
        requires: dict[object, VariableABC],
        resumer: "RawResumer",
        override_exception_location: bool = True,
    ) -> None:
        Waiter.__init__(self, queue_scheduler)
        self.resolver = resolver
        self.requires = requires
        self.resumer = resumer
        self.override_exception_location: bool = override_exception_location
        for r in requires.values():
            self.waitfor(r)
        self.ready(self)

    def execute(self) -> None:
        try:
            self.resumer.resume(self.requires, self.resolver, self.queue)
        except RuntimeException as e:
            if self.override_exception_location:
                e.set_statement(self.resumer)
                e.location = self.resumer.location
            raise e
        self.done = True


"""
Resolution

 - lexical scope (module it is in, then parent modules)
   - handled in NS and direct_lookup on XU
 - dynamic scope (entity, for loop,...)
   - resolvers
   - lex scope as arg or root

resolution
 - FQN:  to lexical scope (via parents) -> to its  NS (imports) -> to target NS -> directLookup
 - N: to resolver -> to parents -> to NS (lex root) -> directlookup

"""

Typeorvalue = Union[Type, ResultVariable]


class Resolver:
    __slots__ = ("namespace", "dataflow_graph")

    def __init__(self, namespace: Namespace, enable_dataflow_graph: bool = False) -> None:
        self.namespace = namespace
        self.dataflow_graph: Optional[DataflowGraph] = DataflowGraph(self) if enable_dataflow_graph else None

    def lookup(self, name: str, root: Optional[Namespace] = None) -> Typeorvalue:
        # override lexial root
        # i.e. delegate to parent, until we get to the root, then either go to our root or lexical root of our caller
        if root is not None:
            ns = root
        else:
            ns = self.namespace

        return ns.lookup(name)

    def get_root_resolver(self) -> "Resolver":
        return self

    def for_namespace(self, namespace: Namespace) -> "Resolver":
        return NamespaceResolver(self, namespace)

    def get_dataflow_node(self, name: str) -> "dataflow.AssignableNodeReference":
        try:
            result_variable: Typeorvalue = self.lookup(name)
            assert isinstance(result_variable, ResultVariable)
            return result_variable.get_dataflow_node()
        except NotFoundException:
            # This block is only executed if the model contains a reference to an undefined variable.
            # Since we don't know in which scope it should be defined, we assume top scope.
            root_graph: Optional[DataflowGraph] = self.get_root_resolver().dataflow_graph
            assert root_graph is not None
            return root_graph.get_own_variable(name)


class VariableResolver(Resolver):
    """
    Resolver that resolves a single variable to a value, and delegates the rest to its parent resolver.
    """

    __slots__ = (
        "parent",
        "name",
        "variable",
    )

    def __init__(self, parent: Resolver, name: str, variable: ResultVariable[T]) -> None:
        self.parent: Resolver = parent
        self.name: str = name
        self.variable: ResultVariable[T] = variable
        self.dataflow_graph = (
            DataflowGraph(self, parent=self.parent.dataflow_graph) if self.parent.dataflow_graph is not None else None
        )

    def lookup(self, name: str, root: Optional[Namespace] = None) -> Typeorvalue:
        if root is None and name == self.name:
            return self.variable
        return self.parent.lookup(name, root)

    def get_root_resolver(self) -> "Resolver":
        return self.parent.get_root_resolver()


class NamespaceResolver(Resolver):
    __slots__ = ("parent", "root")

    def __init__(self, parent: Resolver, lecial_root: Namespace) -> None:
        self.parent = parent
        self.root = lecial_root
        self.dataflow_graph: Optional[DataflowGraph] = None
        if parent.dataflow_graph is not None:
            self.dataflow_graph = DataflowGraph(self, parent.dataflow_graph)

    def lookup(self, name: str, root: Optional[Namespace] = None) -> Typeorvalue:
        if root is not None:
            return self.parent.lookup(name, root)
        return self.parent.lookup(name, self.root)

    def for_namespace(self, namespace: Namespace) -> "Resolver":
        return NamespaceResolver(self, namespace)

    def get_root_resolver(self) -> "Resolver":
        return self.parent.get_root_resolver()


class ExecutionContext(Resolver):
    __slots__ = ("block", "slots", "resolver")

    def __init__(self, block: "BasicBlock", resolver: Resolver):
        self.block = block
        self.slots: dict[str, ResultVariable] = {n: ResultVariable() for n in block.get_variables()}
        self.resolver = resolver
        self.dataflow_graph: Optional[DataflowGraph] = None
        if resolver.dataflow_graph is not None:
            self.dataflow_graph = DataflowGraph(self, resolver.dataflow_graph)
            for name, var in self.slots.items():
                node_ref: dataflow.AssignableNodeReference = dataflow.AssignableNode(name).reference()
                var.set_dataflow_node(node_ref)

    def lookup(self, name: str, root: Optional[Namespace] = None) -> Typeorvalue:
        if "::" in name:
            return self.resolver.lookup(name, root)
        if name in self.slots:
            return self.slots[name]
        return self.resolver.lookup(name, root)

    def direct_lookup(self, name: str) -> ResultVariable:
        if name in self.slots:
            return self.slots[name]
        else:
            raise NotFoundException(None, name, "variable %s not found" % name)

    def emit(self, queue: QueueScheduler) -> None:
        self.block.emit(self, queue)

    def get_root_resolver(self) -> Resolver:
        return self.resolver.get_root_resolver()

    def for_namespace(self, namespace: Namespace) -> Resolver:
        return NamespaceResolver(self, namespace)


# also extends locatable
class Instance(ExecutionContext, references.MaybeReference):
    def set_location(self, location: Location) -> None:
        Locatable.set_location(self, location)
        self.locations.append(location)

    def get_location(self) -> Optional[Location]:
        return Locatable.get_location(self)

    location = property(get_location, set_location)

    __slots__ = (
        "_location",
        "resolver",
        "type",
        "sid",
        "implementations",
        "locations",
        "instance_node",
        "dataclass_self",
    )

    def __init__(
        self,
        mytype: "Entity",
        resolver: Resolver,
        queue: QueueScheduler,
        node: Optional["dataflow.InstanceNodeReference"] = None,
    ) -> None:
        Locatable.__init__(self)
        # ExecutionContext, Resolver -> this class only uses it as an "interface", so no constructor call!
        self.resolver = resolver.get_root_resolver()
        self.type = mytype
        self.slots: dict[str, ResultVariable] = {}
        for attr_name in mytype.get_all_attribute_names():
            if attr_name in self.slots:
                # prune duplicates because get_new_result_variable() has side effects
                # don't use set for pruning because side effects drive control flow and set iteration is nondeterministic
                continue
            attribute = mytype.get_attribute(attr_name)
            assert attribute is not None  # Make mypy happy
            self.slots[attr_name] = attribute.get_new_result_variable(self, queue)
        # TODO: this is somewhat ugly. Is there a cleaner way to enforce this constraint
        assert (resolver.dataflow_graph is None) == (node is None)
        self.dataflow_graph: Optional[DataflowGraph] = None
        self.instance_node: Optional[dataflow.InstanceNodeReference] = node
        if self.instance_node is not None:
            self.dataflow_graph = DataflowGraph(self, resolver.dataflow_graph)
            for name, var in self.slots.items():
                var.set_dataflow_node(dataflow.InstanceAttributeNodeReference(self.instance_node.top_node(), name))

        self.slots["self"] = ResultVariable()
        self.slots["self"].set_value(self, None)
        if self.instance_node is not None:
            self_var_node: dataflow.AssignableNodeReference = dataflow.AssignableNode("__self__").reference()
            self_var_node.assign(self.instance_node, self, cast(DataflowGraph, self.dataflow_graph))
            self.slots["self"].set_dataflow_node(self_var_node)

        self.sid = id(self)
        self.implementations: "set[Implementation]" = set()

        self.locations: list[Location] = []
        # may also be a reference to a dataclass
        self.dataclass_self: object | references.Reference[references.RefValue] | None = None

    def get_type(self) -> "Entity":
        return self.type

    def unwrap_reference(self) -> Optional[references.Reference[references.RefValue]]:
        return (
            self.dataclass_self
            if self.dataclass_self is not None and isinstance(self.dataclass_self, references.Reference)
            else None
        )

    def set_attribute(self, name: str, value: object, location: Location, recur: bool = True) -> None:
        if name not in self.slots:
            raise NotFoundException(None, name, f"cannot set attribute with name {name} on type {self.type}")
        try:
            self.slots[name].set_value(value, location, recur)
        except RuntimeException as e:
            raise AttributeException(self, self, name, cause=e)

    def get_attribute(self, name: str) -> ResultVariable:
        try:
            return self.slots[name]
        except KeyError:
            raise NotFoundException(None, name, f"could not find attribute with name: {name} in type {self.type}")

    def __repr__(self) -> str:
        return f"{self.type} {self.sid:02x}"

    def __str__(self) -> str:
        return "{} (instantiated at {})".format(self.type, ",".join([str(location) for location in self.get_locations()]))

    def add_implementation(self, impl: "Implementation") -> bool:
        if impl in self.implementations:
            return False
        self.implementations.add(impl)
        return True

    def final(self, excns: list[CompilerException]) -> None:
        """
        The object should be complete, freeze all attributes
        """
        if len(self.implementations) == 0:
            excns.append(RuntimeException(self, "Unable to select implementation for entity %s" % self.type.name))

        for k, v in self.slots.items():
            if not v.is_ready():
                if v.can_get():
                    v.freeze()
                else:
                    attr = self.type.get_attribute(k)
                    assert attr is not None  # Make mypy happy
                    if isinstance(attr, ast.attribute.RelationAttribute) and attr.is_multi():
                        low = attr.low
                        # none for list attributes
                        # list for n-ary relations
                        length = 0 if v.value is None else len(v.value)
                        excns.append(
                            inmanta.ast.UnsetException(
                                "The object %s is not complete: attribute %s (%s) requires %d values but only %d are set"
                                % (self, k, attr.location, low, length),
                                self,
                                attr,
                            )
                        )
                    else:
                        excns.append(
                            inmanta.ast.UnsetException(
                                f"The object {self} is not complete: attribute {k} ({attr.location}) is not set",
                                self,
                                attr,
                            )
                        )

    def verify_done(self) -> bool:
        for v in self.slots.values():
            if not v.can_get():
                return False
        return True

    def get_locations(self) -> list[Location]:
        return self.locations



"""
Copyright 2014 Omer Gertel
Copyright 2025 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com

This code was originally developed by Omer Gertel, as a python port of the core portion of a
[Java Metrics library by Coda Hale](http://metrics.dropwizard.io/)

It was vendored into the inmanta source tree as the original was no longer maintained.
"""

import math
import time
from threading import Lock

from .. import Clock
from ..stats.samples import DEFAULT_ALPHA, DEFAULT_SIZE, ExpDecayingSample, Sample
from ..stats.snapshot import Snapshot


class Histogram(object):
    """
    A metric which calculates the distribution of a value.
    """

    counter: float  # Would be expected to be int?
    max: float
    min: float
    sum: float
    var: tuple[float, float]
    sample: Sample

    def __init__(
        self, size: int = DEFAULT_SIZE, alpha: float = DEFAULT_ALPHA, clock: Clock = time, sample: Sample | None = None
    ) -> None:
        """
        Creates a new instance of a L{Histogram}.
        """
        super(Histogram, self).__init__()
        self.lock = Lock()
        self.clock = clock
        if sample is None:
            sample = ExpDecayingSample(size, alpha, clock)
        self.sample = sample
        self.clear()

    def add(self, value: float) -> None:
        """
        Add value to histogram

        :type value: float
        """
        with self.lock:
            self.sample.update(value)
            self.counter = self.counter + 1
            self.max = value if value > self.max else self.max
            self.min = value if value < self.min else self.min
            self.sum = self.sum + value
            self._update_var(value)

    def clear(self) -> None:
        "reset histogram to initial state"
        with self.lock:
            self.sample.clear()
            self.counter = 0.0
            self.max = -2147483647.0
            self.min = 2147483647.0
            self.sum = 0.0
            self.var = (-1.0, 0.0)

    def get_count(self) -> float:
        "get current value of counter"
        return self.counter

    def get_sum(self) -> float:
        "get current sum"
        return self.sum

    def get_max(self) -> float:
        "get current maximum"
        return self.max

    def get_min(self) -> float:
        "get current minimum"
        return self.min

    def get_mean(self) -> float:
        "get current mean"
        if self.counter > 0:
            return self.sum / self.counter
        return 0

    def get_stddev(self) -> float:
        "get current standard deviation"
        if self.counter > 0:
            return math.sqrt(self.get_var())
        return 0

    def get_var(self) -> float:
        "get current variance"
        if self.counter > 1:
            return self.var[1] / (self.counter - 1)
        return 0

    def get_snapshot(self) -> Snapshot:
        "get snapshot instance which holds the percentiles"
        return self.sample.get_snapshot()

    def _update_var(self, value: float) -> None:
        old_m, old_s = self.var
        new_m, new_s = (0.0, 0.0)
        if old_m == -1:
            new_m = value
        else:
            new_m = old_m + ((value - old_m) / self.counter)
            new_s = old_s + ((value - old_m) * (value - new_m))
        self.var = (new_m, new_s)



"""
Copyright 2019 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import logging
import uuid
from typing import Optional, cast

import asyncpg

from inmanta import data
from inmanta.data import model
from inmanta.protocol import handle, methods, methods_v2
from inmanta.protocol.exceptions import Conflict, NotFound, ServerError
from inmanta.server import (
    SLICE_AUTOSTARTED_AGENT_MANAGER,
    SLICE_DATABASE,
    SLICE_PROJECT,
    SLICE_RESOURCE,
    SLICE_TRANSPORT,
    protocol,
)
from inmanta.server.agentmanager import AutostartedAgentManager
from inmanta.server.services.resourceservice import ResourceService
from inmanta.types import Apireturn, JsonType

LOGGER = logging.getLogger(__name__)


class ProjectService(protocol.ServerSlice):
    """Slice with project and environment management"""

    autostarted_agent_manager: AutostartedAgentManager
    resource_service: ResourceService

    def __init__(self) -> None:
        super().__init__(SLICE_PROJECT)

    def get_dependencies(self) -> list[str]:
        return [SLICE_DATABASE, SLICE_RESOURCE, SLICE_AUTOSTARTED_AGENT_MANAGER]

    def get_depended_by(self) -> list[str]:
        return [SLICE_TRANSPORT]

    async def prestart(self, server: protocol.Server) -> None:
        await super().prestart(server)
        self.autostarted_agent_manager = cast(AutostartedAgentManager, server.get_slice(SLICE_AUTOSTARTED_AGENT_MANAGER))
        self.resource_service = cast(ResourceService, server.get_slice(SLICE_RESOURCE))

    # v1 handlers
    @handle(methods.create_project)
    async def create_project(self, name: str, project_id: Optional[uuid.UUID]) -> Apireturn:
        return 200, {"project": (await self.project_create(name, project_id)).model_dump()}

    @handle(methods.delete_project, project_id="id", api_version=1)
    async def delete_project(self, project_id: uuid.UUID) -> Apireturn:
        await self.project_delete(project_id)
        return 200

    @handle(methods.modify_project, project_id="id")
    async def modify_project(self, project_id: uuid.UUID, name: str) -> Apireturn:
        return 200, {"project": (await self.project_modify(project_id, name)).model_dump()}

    @handle(methods.list_projects)
    async def list_projects(self) -> Apireturn:
        project_list: list[JsonType] = [x.model_dump() for x in await self.project_list()]
        for project in project_list:
            project["environments"] = [x["id"] for x in project["environments"]]
        return 200, {"projects": project_list}

    @handle(methods.get_project, project_id="id")
    async def get_project(self, project_id: uuid.UUID) -> Apireturn:
        project_model = (await self.project_get(project_id)).model_dump()
        project_model["environments"] = [e.id for e in await data.Environment.get_list(project=project_id)]
        return 200, {"project": project_model}

    # v2 handlers
    @handle(methods_v2.project_create)
    async def project_create(self, name: str, project_id: Optional[uuid.UUID]) -> model.Project:
        if project_id is None:
            project_id = uuid.uuid4()

        try:
            project = data.Project(id=project_id, name=name)
            await project.insert()
        except asyncpg.exceptions.UniqueViolationError:
            raise ServerError(f"A project with name {name} already exists.")

        return project.to_dto()

    @handle(methods_v2.project_delete, project_id="id", api_version=2)
    async def project_delete(self, project_id: uuid.UUID) -> None:
        project = await data.Project.get_by_id(project_id)
        if project is None:
            raise NotFound("The project with given id does not exist.")

        environments = await data.Environment.get_list(project=project.id)
        if len(environments) > 0:
            raise Conflict(
                f"Cannot remove the project `{project_id}` because it still contains some environments: "
                f"{','.join([str((env.name, str(env.id))) for env in environments])}"
            )

        await project.delete()

    @handle(methods_v2.project_modify, project_id="id")
    async def project_modify(self, project_id: uuid.UUID, name: str) -> model.Project:
        try:
            project = await data.Project.get_by_id(project_id)
            if project is None:
                raise NotFound("The project with given id does not exist.")

            await project.update_fields(name=name)

            return project.to_dto()

        except asyncpg.exceptions.UniqueViolationError:
            raise ServerError(f"A project with name {name} already exists.")

    @handle(methods_v2.project_list)
    async def project_list(self, environment_details: bool = False) -> list[model.Project]:
        project_list = []

        for project in await data.Project.get_list(order_by_column="name", order="ASC"):
            project_model = project.to_dto()
            project_model.environments = [
                e.to_dto()
                for e in await data.Environment.get_list(
                    project=project.id, details=environment_details, order_by_column="name", order="ASC"
                )
            ]

            project_list.append(project_model)

        return project_list

    @handle(methods_v2.project_get, project_id="id")
    async def project_get(self, project_id: uuid.UUID, environment_details: bool = False) -> model.Project:
        project = await data.Project.get_by_id(project_id)

        if project is None:
            raise NotFound("The project with given id does not exist.")

        project_model = project.to_dto()
        project_model.environments = [
            e.to_dto() for e in await data.Environment.get_list(project=project.id, details=environment_details)
        ]

        return project_model
